# Setting asym calculation histograms window
# Derek Fujimoto
# August 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import webbrowser
import textwrap
import os
import logging

# ========================================================================== #
class popup_set_histograms(object):
    """
        Popup window for setting asymmetry calculation histograms. 
    """

    # ====================================================================== #
    def __init__(self, parent):
        self.parent = parent
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(parent.mainframe)
        self.win.title('Set Histograms')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        
        # icon
        self.parent.set_icon(self.win)
        
        # Key bindings
        self.win.bind('<Return>', self.set)             
        self.win.bind('<KP_Enter>', self.set)
        
        # labels 
        title_label = ttk.Label(frame, text=\
                'Set histogram names for asymmetry calcultions.')
                
        explanation_frame = ttk.Frame(frame, relief='sunken', pad=5)
        explanation_label = ttk.Label(explanation_frame, 
            text=textwrap.dedent("""\
            If we wished to do a simple asymmetry calculation in the form of 
                                    
                                    (F-B)/(F+B)
            
            for each helicity, then 
                                       
                                    'F+, F-, B+, B-'
                                       
            for alpha diffusion calculations append the two alpha counters
            
                                 'R+, R-, L+, L-, A+, A-
            
            for alpha tagged calculations do the following
            
                     'R+, R-, L+, L-, TR+, TR-, TL+, TL-, nTR+, nTR-, nTL+, nTL-'
                    
                where TR is the right counter tagged (coincident) with alphas, 
                      TL is the left  counter tagged with alphas, 
                     nTR is the right counter tagged with !alphas (absence of), 
                     nLR is the right counter tagged with !alphas
                     
            (note the lack of spaces in each list)
            """), \
            pad=5, justify=LEFT, font=('Courier'))
        
        self.text = StringVar()
        self.text.set(self.parent.hist_select)
        self.entry = Entry(frame, textvariable=self.text, width=75, justify=CENTER)
        
        try:
            if hasattr(parent.fileviewer, 'data'): 
                s = list(parent.fileviewer.data.hist.keys())
            else:
                run = list(parent.fetch_files.data.keys())[0]
                s = list(parent.fetch_files.data[run].hist.keys())
        except (IndexError, AttributeError):
            self.logger.exception('No files fetched or viewed')
            s = 'No files fetched or viewed'
        else:
            s.sort()
            s = ''.join(["'%s', "%i for i in s])
            s = s[:-2]
            s = textwrap.wrap(s, 75)
            s = '\n'.join(s)
            
        detected_label = ttk.Label(frame, text='The detected histogram titles are:')
        detected_label2 = ttk.Label(frame, text=s, font=('Courier'), justify=CENTER)
        
        # grid 
        r = 0
        title_label.grid(column=0, row=r, columnspan=3, sticky=(W, E)); r+=1
        explanation_frame.grid(column=0, row=r, columnspan=3, sticky=(W, E), pady=10); r+=1
        explanation_label.grid(column=0, row=0, sticky=(W, E))
        detected_label.grid(column=0, row=r, sticky=(W, E), columnspan=3, pady=20); r+= 1
        detected_label2.grid(column=0, row=r, sticky=(W, E), columnspan=3); r+= 1
        self.entry.grid(column=0, row=r, columnspan=3, pady=20); r+=1
          
        # add buttons
        set_button = ttk.Button(frame, text='Set', command=self.set)
        close_button = ttk.Button(frame, text='Cancel', command=self.cancel)
        help_button = ttk.Button(frame, text='Help', command=self.help)
        set_button.grid(column=0,  row=r)
        help_button.grid(column=1, row=r)
        close_button.grid(column=2, row=r)
            
        # grid frame
        frame.grid(column=0, row=0)
        
        self.logger.debug('Initialization success. Starting mainloop.')

    # ====================================================================== #
    def help(self):
        self.logger.info('Opening help')
        p = os.path
        webbrowser.open(p.split(p.abspath(p.realpath(__file__)))[0]+'/help.html')

    # ====================================================================== #
    def set(self, *args):
        """Set entered values"""
        self.parent.hist_select = self.entry.get()
        self.logger.info('Set histogram selection to "%s"', self.parent.hist_select)
        self.win.destroy()
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()



# Set up default fitting routines. 
# Derek Fujimoto
# Aug 2018

import bfit.fitting.functions as fns
from bfit.fitting.decay_31mg import fa_31Mg
from functools import partial
import collections
import numpy as np
import bdata as bd

class fitter(object):
    """
        Fitter base class for default minimizers
    """

    # needed to tell users what routine this is
    __name__ = 'base'
    
    # Define possible fit functions for given run modes
    function_names = {  '20':('Exp', 'Bi Exp', 'Str Exp'), 
                        '2h':('Exp', 'Bi Exp', 'Str Exp'), 
                        '1f':('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz'), 
                        '1w':('Lorentzian', 'Gaussian', 'BiLorentzian', ), 
                        '1n':('Lorentzian', 'Gaussian', 'BiLorentzian', ), 
                        '2e':('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz')}
     
    # Define names of fit parameters:
    param_names = {     'Exp'       :('1_T1', 'amp'), 
                        'Bi Exp'    :('1_T1', '1_T1b', 'fraction_b', 'amp'), 
                        'Str Exp'   :('1_T1', 'beta', 'amp'), 
                        'Lorentzian':('peak', 'fwhm', 'height', 'baseline'), 
                        'BiLorentzian':('peak', 'fwhmA', 'heightA', 
                                               'fwhmB', 'heightB', 'baseline'), 
                        'QuadLorentz':('nu_0', 'nu_q', 'efgAsym', 'efgTheta', 'efgPhi', 
	                               'amp0', 'amp1', 'amp2', 'amp3', 'fwhm', 
                                       'baseline'), 
                        'Gaussian'  :('mean', 'sigma', 'height', 'baseline'), }

    # nice parameter names for labels - must be one to one unique for later inversion
    pretty_param = {'1_T1':             r'$1 / T_1$ (s$^{-1}$)', 
                    'amp':              'Amplitude', 
                    'fraction_b':       r'Fraction of $T_1^{(b)}$', 
                    '1_T1b':            r'$1/T_1^{(b)}$ (s$^{-1}$)', 
                    'beta':             r'$\beta$', 
                    'fwhm':             'FWHM (%s)', 
                    'height':           'Height', 
                    'baseline':         'Baseline', 
                    'peak':             'Peak (%s)', 
                    'fwhmA':            'FWHM$_A$ (%s)', 
                    'fwhmB':            'FWHM$_B$ (%s)', 
                    'heightA':          'Height$_A$', 
                    'heightB':          'Height$_B$', 
                    'nu_0':             r'$\nu_0$ (%s)', 
                    'nu_q':             r'$\nu_q$ (%s)', 
                    'efgAsym':          r'$\eta$', 
                    'efgTheta':         r'Principle Axes $\theta$ (rad)', 
                    'efgPhi':           r'Principle Axes $\phi$ (rad)', 
                    'amp0':             'Amplitude$_0$', 
                    'amp1':             'Amplitude$_1$', 
                    'amp2':             'Amplitude$_2$', 
                    'amp3':             'Amplitude$_3$', 
                    'fmwh':             'FWHM (%s)', 
                    'mean':             r'$\langle\nu\rangle$ (%s)', 
                    'sigma':            r'$\sigma$ (%s)', 
                    'Beta-Avg 1/<T1>':  r'$1/\langle T_1\rangle_\beta$ (s$^{-1}$)', 
                    'B0 Field (T)':     r'$B_0$ Field (T)', 
                    }

    # dictionary of initial parameters
    par_values = {}
    fn_list = {}
    epsilon = 1e-9  # for fixing parameters

    # define list of ok run modes 
    valid_asym_modes = ('c', 'p', 'n', 'sl_c', 'dif_c', )
    
    # probe species spin numbers
    spin = {'Li8':2,    # radioactive daughters: none
            'Li9':1.5,  # radioactive daughters: Be8
            'Li11':0.5, # radioactive daughters: Be11, He6, Be10 
            'Be11':0.5, # radioactive daughters: none 
            'F20':2,    # radioactive daughters: none
            'Mg31':0.5, # radioactive daughters: Al31, Si31 
            'Mg29':1.5, # radioactive daughters: Al29
            'Ac230':1,  # radioactive daughters: none
            'Ac232':1,  # radioactive daughters: unknown
            'Ac234':1,  # radioactive daughters: unkonwn
            }

    # ======================================================================= #
    def __init__(self, keyfn, probe_species='Li8'):
        """
            keyfn:          function takes as input bdata or bjoined or bmerged 
                            object, returns string corresponding to unique id of 
                            that object
            probe_species: one of the keys in the bdata.life dictionary.
        """
        self.keyfn = keyfn
        self.probe_species = probe_species
        
    # ======================================================================= #
    def __call__(self, fn_name, ncomp, data_list, hist_select, asym_mode, xlims):
        """
            Fitting controller. 
            
            fn_name: name of function to fit
            ncomp : number of components to incude (2 = biexp, for example)
            data_list: list of [[bdata object, pdict, doptions], ]
            
                where pdict = {par:(init val,   # initial guess
                                    bound_lo,   # lower fitting bound
                                    bound_hi,   # upper fitting bound
                                    is_fixed,   # boolean, fix value?
                                    is_shared,  # boolean, share value globally?
                                   )
                              }
                where doptions = {  'omit':str,     # bins to omit in 1F calcs
                                    'rebin':int,    # rebinning factor
                                    'group':int,    # fitting group
                                 }
                                 
            hist_select: string for selection of histograms
            asym_mode:  input for asymmetry calculation type 
                                               c: combined helicity
                            h: split helicity
                            p: positive helicity
                            n: negative helicity
                            
                        For 2e mode, prefix with:
                            sl_: combined timebins using slopes
                            dif_: combined timebins using differences
                            raw_: raw time-resolved
                            
                            ex: sl_c or raw_h or dif_c
            xlims: fit subrange of x axis
            returns dictionary of {runid: [[par_names], [par_values], [par_errors], 
                                          [chisquared], [fitfunction pointers]]}
                                   and global chisquared
        """

        # check ncomponents
        if ncomp < 1:
            raise RuntimeError('ncomp needs to be >= 1')
            
        asym_mode = asym_mode.replace('2e_', '')
            
        # parameter names
        keylist = self.gen_param_names(fn_name, ncomp)
        npar = len(keylist)
        
        # gather list of data to fit 
        fn = []
        bdata_list = []
        p0 = []
        bounds = []
        omit = []
        rebin = []
        sharelist = np.zeros(npar, dtype=bool)
        fixedlist = []
        
        for data in data_list:
            
            # split data list into parts
            dat = data[0]
            pdict = data[1]
            doptions = data[2]
            
            # probe lifetime
            life = bd.life[self.probe_species]
            
            # get fitting function for 20 and 2h
            if dat.mode in ['20', '2h']: 
                pulse = dat.get_pulse_s()
                
                # fit function
                fn1 = self.get_fn(fn_name, ncomp, pulse, life)
                
                # add corrections for probe daughters
                if self.probe_species == 'Mg31':
                    fn.append(lambda x, *par : fa_31Mg(x, pulse)*fn1(x, *par))
                else:
                    fn.append(fn1)
                
            # 1f functions
            else:                       
                fn.append(self.get_fn(fn_name, ncomp, -1, life))
            
            # get bdata objects
            bdata_list.append(dat)
            
            # get initial parameters
            p0.append(tuple(pdict[k][0] for k in keylist))
            
            # get fitting bounds
            bound = [[], []]
            shlist = []
            fixed = []
            for k in keylist:
                
                # bounds
                bound[0].append(pdict[k][1])
                bound[1].append(pdict[k][2])
                
                # fixed
                fixed.append(pdict[k][3])    
                    
                # sharelist
                shlist.append(pdict[k][4])
                
            bounds.append(bound)
            fixedlist.append(fixed)
            
            # look for any sharelist
            sharelist += np.array(shlist)
            
            # rebin and omit
            try:
                rebin.append(doptions['rebin'])
            except KeyError:
                rebin.append(1)
                
            try:
                omit.append(doptions['omit'])
            except KeyError:
                omit.append('')                
            
        # fit data
        kwargs = {'p0':p0, 'bounds':bounds}
        pars, stds_l, stds_h, covs, chis, gchi = self._do_fit(
                                        bdata_list, 
                                        fn, 
                                        omit, 
                                        rebin, 
                                        sharelist, 
                                        hist_select=hist_select, 
                                        asym_mode=asym_mode, 
                                        fixed=fixedlist, 
                                        xlims=xlims, 
                                        parnames=keylist, 
                                        **kwargs)
        
        # collect results
        if not isinstance(chis, collections.Iterable):   # single run
            d = bdata_list[0]
            return ({self.keyfn(d):[keylist, pars, stds_l, stds_h, chis]}, gchi)
        else:                                           # multiple runs    
            return ({self.keyfn(d):[keylist, p, sl, sh, c] 
                for d, p, sl, sh, c in zip(bdata_list, pars, stds_l, stds_h, chis)}, gchi)

    # ======================================================================= #
    def get_fit_fn(self, fn_name, ncomp, data_list):
        
        fn = []
        
        for data in data_list:
            
            # split data list into parts
            dat = data[0]
            pdict = data[1]
            doptions = data[2]
            
            # probe lifetime
            life = bd.life[self.probe_species]
            
            # get fitting function for 20 and 2h
            if dat.mode in ['20', '2h']: 
                pulse = dat.get_pulse_s()
                
                # fit function
                fn1 = self.get_fn(fn_name, ncomp, pulse, life)
                
                # add corrections for probe daughters
                if self.probe_species == 'Mg31':
                    fn.append(lambda x, *par : fa_31Mg(x, pulse)*fn1(x, *par))
                else:
                    fn.append(fn1)
                
            # 1f functions
            else:                       
                fn.append(self.get_fn(fn_name, ncomp, -1, life))

        return ({self.keyfn(d[0]):f for d, f in zip(data_list, fn)})

    # ======================================================================= #
    def gen_param_names(self, fn_name, ncomp):
        """
            Make a list of the parameter names based on the number of components.
            
            fn_name: name of function (should match those in param_names)
            ncomp: number of components
            
            return (names)
        """
        
        # get names
        names_orig = self.param_names[fn_name]
        
        # special case of one component
        if ncomp == 1: 
            return names_orig
        
        # multicomponent: make copies of everything other than the baselines
        names = []
        for c in range(ncomp): 
            for n in names_orig:
                if 'base' in n: continue
                names.append(n+'_%d' % c)
                
        if 'base' in names_orig[-1]:
            names.append(names_orig[-1])
        
        return tuple(names)
        
    # ======================================================================= #
    def gen_init_par(self, fn_name, ncomp, bdataobj, asym_mode='combined'):
        """Generate initial parameters for a given function.
        
            fname: name of function. Should be the same as the param_names keys
            ncomp: number of components
            bdataobj: a bdata object representative of the fitting group. 
            asym_mode: what kind of asymmetry to fit
            
            Set and return dictionary of initial parameters. 
                {par_name:par_value}
        """
        
        # asym_mode un-used types 
        if asym_mode in (   'h',           # Split Helicity          
                            'hm',          # Matched Helicity
                            'hs',          # Shifted Split          
                            'cs',          # Shifted Combined      
                            'hp',          # Matched Peak Finding
                            'r',           # Raw Scans            
                            'rhist',       # Raw Histograms          
                            '2e_raw_c',    # Combined Hel Raw       
                            '2e_raw_h',    # Split Hel Raw          
                            '2e_sl_h',     # Split Hel Slopes
                            '2e_dif_h',    # Split Hel Diff           
                            'ad',          # Alpha Diffusion  
                            "at_c",        # Combined Hel (Alpha Tag)
                            "at_h",        # Split Hel (Alpha Tag) 
                            "nat_c",       # Combined Hel (!Alpha Tag)
                            "nat_h",       # Split Hel (!Alpha Tag)
                        ):
            errmsg = "Asymmetry calculation type not implemented for fitting"
            raise RuntimeError(errmsg)
        
        # get asymmetry
        x, a, da = bdataobj.asym(asym_mode)
        
        # set pulsed exp fit initial parameters
        if fn_name in ('Exp', 'Bi Exp', 'Str Exp'):
            # ampltitude average of first 5 bins
            amp = abs(np.mean(a[0:5])/ncomp)
            
            # T1: time after beam off to reach 1/e
            idx = int(bdataobj.ppg.beam_on.mean)
            beam_duration = x[idx]
            amp_beamoff = a[idx]
            target = amp_beamoff/np.exp(1)
            
            x_target = x[np.sum(a>target)]
            T1 = abs(x_target-beam_duration)
            
            # bounds and amp
            if asym_mode == 'n':
                amp_bounds = (-np.inf, np.inf)
                amp = -amp
            else:
                amp_bounds = (0, np.inf)
            
            # set values
            if fn_name == 'Exp':
                par_values = {  '1_T1':(1./T1, 0, np.inf, False), 
                                'amp':(amp, *amp_bounds, False), 
                             }
                            
            elif fn_name == 'Bi Exp':
                par_values = {  '1_T1':(1./T1, 0, np.inf, False), 
                                '1_T1b':(10./T1, 0, np.inf, False), 
                                'fraction_b':(0.5, 0, 1, False), 
                                'amp':(amp, *amp_bounds, False), 
                             }
            
            elif fn_name == 'Str Exp':
                par_values = {  '1_T1':(1./T1, 0, np.inf, False), 
                                'beta':(0.5, 0, 1, False),
                                'amp':(amp, *amp_bounds, False), 
                             }
                            
        # set time integrated fit initial parameters
        elif fn_name in ('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz'):
            
            # get peak asym value
            amin = min(a[a!=0])
            
            peak = x[np.where(a==amin)[0][0]]
            base = np.mean(a[:5])
            height = abs(base-amin)
            width = 2*abs(peak-x[np.where(a<amin+height/2)[0][0]])
            
            # bounds
            if asym_mode == 'n':
                height_bounds = (-np.inf, 0)
            else:
                height_bounds = (0, np.inf)
            
            # set values (value, low bnd, high bnd, fixed)
            if fn_name == 'Lorentzian':	
                par_values = {'peak':(peak, min(x), max(x), False), 
                              'fwhm':(width, 0, np.inf, False), 
                              'height':(height, *height_bounds, False), 
                              'baseline':(base, -np.inf, np.inf, False)
                             }
            elif fn_name == 'Gaussian':
                par_values = {'mean':(peak, min(x), max(x), False), 
                              'sigma':(width, 0, np.inf, False), 
                              'height':(height, *height_bounds, False), 
                              'baseline':(base, -np.inf, np.inf, False)
                              }
            elif fn_name == 'BiLorentzian':
                par_values = {'peak':(peak, min(x), max(x), False), 
                              'fwhmA':(width*10, 0, np.inf, False), 
                              'heightA':(height/10, *height_bounds, False), 
                              'fwhmB':(width, 0, np.inf, False), 
                              'heightB':(height*9/10, *height_bounds, False), 
                              'baseline':(base, -np.inf, np.inf, False)
                             }
            elif fn_name == 'QuadLorentz':
                
                dx = max(x)-min(x)
                par_values = {'nu_0':((max(x)+min(x))/2, min(x), max(x), False), 
                              'nu_q':(dx/12, 0, dx, False), 
                              'efgAsym':(0, 0, 1, True), 
                              'efgTheta':(0, 0, 2*np.pi, True), 
                              'efgPhi':(0, 0, 2*np.pi, True), 
                              'amp0':(height, height*0.1, np.inf, False), 
                              'amp1':(height, height*0.1, np.inf, False), 
                              'amp2':(height, height*0.1, np.inf, False), 
                              'amp3':(height, height*0.1, np.inf, False), 
                              'fwhm':(dx/10, 0, dx, False), 
                              'baseline':(base, -np.inf, np.inf, False)
                             }
         
        else:
            raise RuntimeError('Bad function name.')
        
        # do multicomponent
        par_values2 = {}
        if ncomp > 1: 
            for c in range(ncomp): 
                for n in par_values.keys():
                    if 'baseline' not in n:
                        par_values2[n+'_%d' % c] = par_values[n]
                    else:
                        par_values2[n] = par_values[n]
        else:
            par_values2 = par_values
            
        return par_values2
        
    # ======================================================================= #
    def get_fn(self, fn_name, ncomp=1, pulse_len=-1, lifetime=-1):
        """
            Get the fitting function used.
            
                fn_name: string of the function name users will select. 
                ncomp: number of components, ex if 2, then return exp+exp
                pulse_len: duration of beam on in s
                lifetime: lifetime of probe in s
            
            Returns python function(x, *pars)
        """
        
        # set fitting function
        if fn_name == 'Lorentzian':
            fn =  fns.lorentzian
            self.mode=1
        elif fn_name == 'BiLorentzian':
            fn =  fns.bilorentzian
            self.mode=1
        elif fn_name == 'QuadLorentz':
            fn =  lambda freq, nu_0, nu_q, eta, theta, phi, \
            		amp0, amp1, amp2, amp3, fwhm: \
                   	fns.quadlorentzian(freq, nu_0, nu_q, eta, theta, phi, \
		            		amp0, amp1, amp2, amp3, \
                		   	fwhm, fwhm, fwhm, fwhm, \
                		   	I=self.spin[self.probe_species])
            self.mode=1
        elif fn_name == 'Gaussian':
            fn =  fns.gaussian
            self.mode=1
        elif fn_name == 'Exp':
            fn =  fns.pulsed_exp(lifetime, pulse_len)
            self.mode=2
        elif fn_name == 'Bi Exp':
            fn =  fns.pulsed_biexp(lifetime, pulse_len)
            self.mode=2
        elif fn_name == 'Str Exp':
            fn =  fns.pulsed_strexp(lifetime, pulse_len)
            self.mode=2
        else:
            raise RuntimeError('Fitting function not found.')
        
        # Make final function based on number of components
        fnlist = [fn]*ncomp
        
        if self.mode == 1:
            fnlist.append(lambda x, b: b)
        fn = fns.get_fn_superpos(fnlist)
        
        return fn
        
        
        
        
        



# move the working plot to the top and focus on it
# Derek Fujimoto
# July 2019 

import matplotlib.pyplot as plt

# bring window to front
def raise_window():
    wm = plt.get_current_fig_manager()    
    wm.canvas.get_tk_widget().focus_force()
    wm.window.lift()




# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox

import logging, re, os, warnings
import weakref as wref
import numpy as np
import pandas as pd

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window
import bfit.backend.colors as colors

# ========================================================================== #
class template_fit_popup(object):
    """
        Base class for fitting popup windows 
        
        bfit
        fittab
        logger
        
        entry:              Text, text entry for user
        entry_label:        Label, instructions above input text box
        input_fn_text:      string, text defining input functions
        left_frame:         Frame, for additional details 
        new_par:            dataframe, index: parnames, columns: p0, blo, bhi, res, err
        
        output_par_text     text, detected parameter names
        output_text         dict, keys: p0, blo, bhi, res, err, value: tkk.Text objects
       
        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text
       
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """
    
    # default parameter values on new parameter
    default_parvals = { 'p0':1, 
                        'blo':-np.inf, 
                        'bhi':np.inf, 
                        'res':np.nan, 
                        'err':np.nan}
    
    window_title = 'Base class popup window'

    # ====================================================================== #
    def __init__(self, bfit, input_fn_text='', output_par_text='', output_text=''):
        
        self.bfit = bfit
        self.fittab = bfit.fit_files
        
        self.input_fn_text = input_fn_text
        self.output_par_text_val = output_par_text
        
        if not output_text:
            self.output_text_val = {}
        else:
            self.output_text_val = output_text
    
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title(self.window_title)
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        left_frame = ttk.Frame(frame)
        right_frame = ttk.Frame(frame)

        # set icon
        self.bfit.set_icon(self.win)

        # Key bindings
        self.win.bind('<Control-Key-Return>', self.do_fit)
        self.win.bind('<Control-Key-KP_Enter>', self.do_fit)
        
        # Text entry
        entry_frame = ttk.Frame(right_frame, relief='sunken', pad=5)
        self.entry_label = ttk.Label(entry_frame, justify=LEFT, 
                                text='')
        self.entry = Text(entry_frame, width=60, height=13, state='normal')
        self.entry.bind('<KeyRelease>', self.get_input)
        scrollb = Scrollbar(entry_frame, command=self.entry.yview)
        self.entry['yscrollcommand'] = scrollb.set
        
        # Insert default text
        self.entry.insert('1.0', self.input_fn_text.strip())
        
        # text for output
        output_frame = ttk.Frame(right_frame, relief='sunken', pad=5)
        output_head1_label = ttk.Label(output_frame, text='Par Name')
        output_head2_label = ttk.Label(output_frame, text='p0')
        output_head3_label = ttk.Label(output_frame, text='Bounds')
        output_head4_label = ttk.Label(output_frame, text='Result')
        output_head5_label = ttk.Label(output_frame, text='Error (-)')
        output_head6_label = ttk.Label(output_frame, text='Error (+)')
        self.output_par_text = Text(output_frame, width=8, height=8)
        self.output_text = {k:Text(output_frame, width=8, height=8, wrap='none')\
                            for k in ('p0', 'blo', 'bhi', 'res', 'err-', 'err+')}
        
        # default starter strings
        if self.output_par_text_val: 
            self.output_par_text.insert('1.0', self.output_par_text_val)
        self.output_par_text.config(state='disabled')
    
        if self.output_text_val: 
            for k in self.output_text_val:
                self.output_text[k].insert('1.0', self.output_text_val[k])

        # disable results
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='disabled', width=12)

        # key bindings and scrollbar
        scrollb_out = Scrollbar(output_frame, command=self.yview)
        self.output_par_text['yscrollcommand'] = scrollb_out.set
        for k in self.output_text:
            self.output_text[k].bind('<KeyRelease>', self.get_result_input)
            self.output_text[k]['yscrollcommand'] = scrollb_out.set
                
        c = 0; r = 0;
        output_head1_label.grid(column=c, row=r);        c+=1;
        output_head2_label.grid(column=c, row=r);        c+=1;
        output_head3_label.grid(column=c, row=r, 
                                columnspan=2);          c+=2;
        output_head4_label.grid(column=c, row=r);        c+=1;
        output_head5_label.grid(column=c, row=r);        c+=1;
        output_head6_label.grid(column=c, row=r);        c+=1;
        
        c = 0; r += 1;
        self.output_par_text.grid(column=c, row=r, sticky=N); c+=1;
        for k in ('p0', 'blo', 'bhi', 'res', 'err-', 'err+'):
            self.output_text[k].grid(column=c, row=r, sticky=N); c+=1;
        scrollb_out.grid(row=r, column=c, sticky='nsew')
        
        # fitting button 
        fit_button = ttk.Button(right_frame, text='Fit', command=self.do_fit)
        
        # gridding
        self.entry_label.grid(column=0, row=0, sticky=W)
        self.entry.grid(column=0, row=1)
        scrollb.grid(row=1, column=1, sticky='nsew')
        
        # grid to frame
        frame.grid(column=0, row=0)
        left_frame.grid(column=0, row=0, sticky=(N, S))
        right_frame.grid(column=1, row=0, sticky=(N, S))
        
        entry_frame.grid(column=0, row=0, sticky=(N, E, W), padx=1, pady=1)
        output_frame.grid(column=0, row=1, sticky=(N, E, W, S), padx=1, pady=1)
        fit_button.grid(column=0, row=2, sticky=(N, E, W), padx=1, pady=1)
        
        # initialize 
        self.new_par = pd.DataFrame(columns=['name', 'p0', 'blo', 'bhi', 'res', 'err-', 'err+']) 
        self.left_frame = left_frame
        
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()
    
    # ====================================================================== #
    def _do_fit(self, text):
        """
            Do the fitting. Called by do_fit_setup().
            
            output: (par, cov, chi, gchi)
        """
        pass
    
    # ====================================================================== #
    def do_fit(self, *args):
        """
            Set up the fit functions and do the fit. 
            Then map the outputs to the proper displays. 
        """
        
        # parse text
        self.do_parse()
        
        # clean input
        text = self.input_fn_text.split('\n')
        text = [t.strip() for t in text if '=' in t]
        
        # check for no input
        if not text:    return
    
        try:
            # do the fit
            out = self._do_fit(text)
        except Exception as errmsg:
            raise errmsg from None
        
        # check if fit was success
        if out is None: 
            return 
        else:
            par, std_l, std_h  = out
        
        # display output for global parameters
        for i, j in enumerate(self.new_par.index):
            self.new_par.loc[j, 'res'] = par[i]
            self.new_par.loc[j, 'err-'] = std_l[i]
            self.new_par.loc[j, 'err+'] = std_h[i]
        self.set_par_text()
        
    # ====================================================================== #
    def do_parse(self, *args):
        """
            Detect new global variables
            returns split lines, new parameter names 
        """
        
        # clean input
        text = self.input_fn_text.split('\n')
        text = [t.strip() for t in text if '=' in t]
        
        # check for no input
        if not text:
            return
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # check for new parameters
        new_par = []
        for eq in eqn:
            lst = re.split('\W+', eq)    # split list non characters
            
            # throw out known things: numbers numpy equations
            delist = []
            for i, l in enumerate(lst):
                
                # check numpy functions
                if l == 'np':
                    delist.append(i)
                    delist.append(i+1)
                    continue
                
                # check integer
                try: 
                    int(l)
                except ValueError:
                    pass
                else:
                    delist.append(i)
                    continue
                
                # check variables
                if l in self.reserved_pars:  
                    delist.append(i)
                    continue
                    
            delist.sort()
            for i in delist[::-1]:
                try:
                    del lst[i]
                except IndexError:  # error raised on incomplete math: ex "np."
                    pass
                
            new_par.append(lst)

        # add result
        new_par = np.unique(np.concatenate(new_par))
        for k in new_par:
            
            # bad input
            if not k: continue
            
            # set defaults
            if k not in self.new_par['name'].values:
                self.new_par = self.new_par.append({'name':k, **self.default_parvals}, 
                                                   ignore_index=True)
        
        # drop results 
        for i, k in zip(self.new_par.index, self.new_par['name']):
            if k not in new_par:
                self.new_par.drop(i, inplace=True)
        
        # set fields
        self.new_par.sort_values('name', inplace=True)
        self.set_par_text()
        
        # logging
        self.logger.info('Parse found constraints for %s, and defined %s', 
                         sorted(defined), 
                         self.new_par['name'].values.tolist())
    
    # ====================================================================== #
    def get_input(self, *args):
        """Get input from text box."""
        self.input_fn_text = self.entry.get('1.0', END)
        self.do_parse()
        
    # ====================================================================== #
    def get_result_input(self, *args):
        """
            Set new_par row to match changes made by user
        """
    
        # get text
        try:
            text = {k : list(map(float, self.output_text[k].get('1.0', END).split('\n')[:-1])) \
                for k in self.output_text}
        # no update if blank
        except ValueError:
            return 
        
        # dataframe it
        try:
            text = pd.DataFrame(text)   
        # bad input
        except ValueError:
            return
        
        # get names of the parameters
        parnames = self.output_par_text.get('1.0', END).split('\n')[:-1]
        
        # update
        par = self.new_par.set_index('name')
        for i, name in enumerate(parnames):
            par.loc[name] = text.iloc[i]
        par.reset_index(inplace=True)
        self.new_par = par
        self.logger.debug('get_result_input: updated new_par')
    
    # ====================================================================== #
    def set_par_text(self):
        """
            Set the textboxes based on stored results in self.newpar
        """
        
        # get strings
        set_par = self.new_par.astype(str)
        
        # round
        numstr = '%'+('.%df' % self.bfit.rounding)
        for k in ('res', 'err-', 'err+'):
            set_par[k] = set_par.loc[:, k].apply(\
                    lambda x : numstr % np.around(float(x), self.bfit.rounding))
        
        # enable setting
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='normal')
        self.output_par_text.config(state='normal')
        
        self.output_par_text.delete('1.0', END)
        self.output_par_text.insert(1.0, '\n'.join(set_par['name']))
        self.output_par_text_val = '\n'.join(set_par['name'])
        
        for k in self.output_text:
            self.output_text[k].delete('1.0', END)
            self.output_text[k].insert(1.0, '\n'.join(set_par[k]))
            self.output_text_val[k] = '\n'.join(set_par[k])
                
        # disable setting
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='disabled')
        self.output_par_text.config(state='disabled')
        
    # ====================================================================== #
    def yview(self, *args):
        """
            Scrollbar for all output text fields
        """
        self.output_par_text.yview(*args)
        for k in self.output_text:
            self.output_text[k].yview(*args)



# Drawing style window
# Derek Fujimoto
# July 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import webbrowser
import logging

# ========================================================================== #
class popup_units(object):
    """
        Popup window for setting redraw period. 
        
        input:      dict:[StringVar, StringVar]
                    keyed by run mode, [conversion #, unit string]
    """

    # ====================================================================== #
    def __init__(self, parent):
        self.parent = parent    # fit_files pointer
        units = parent.units    # units dict from fit_files
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(parent.mainframe)
        self.win.title('Set Units')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        topframe = ttk.Frame(frame, pad=5)

        # icon
        self.parent.set_icon(self.win)

        # Key bindings
        self.win.bind('<Return>', self.set)             
        self.win.bind('<KP_Enter>', self.set)

        # headers
        l_header = ttk.Label(topframe, text='Conversion', pad=5, justify=LEFT)
        l_header2 = ttk.Label(topframe, text='Unit', pad=5, justify=LEFT)

        # grid
        l_header.grid(column=1, row=0)
        l_header2.grid(column=2, row=0)

        # make entries
        self.input = {}    # [value StringVar, unit StringVar]
        r = 1
        for key in units.keys():
            
            l_val = ttk.Label(topframe, text=key, pad=5, justify=LEFT)
            self.input[key] = [StringVar(), StringVar()]
            
            for i in range(2):
                self.input[key][i].set(units[key][i])
                
            entry_val = Entry(topframe, textvariable=self.input[key][0], width=10, justify=RIGHT)        
            entry_uni = Entry(topframe, textvariable=self.input[key][1], width=5, justify=RIGHT)
            
            l_val.grid(column=0, row=r)
            entry_val.grid(column=1, row=r, padx=5)
            entry_uni.grid(column=2, row=r, padx=5)
            r += 1
            
        # make objects: buttons
        set_button = ttk.Button(frame, text='Set', command=self.set)
        close_button = ttk.Button(frame, text='Cancel', command=self.cancel)
        
        # grid
        topframe.grid(column=0, row=0, columnspan=2, pady=10)
        set_button.grid(column=0, row=r, pady=10)
        close_button.grid(column=1, row=r, pady=10)
            
        # grid frame
        frame.grid(column=0, row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def set(self, *args):
        """Set entered values"""     
           
        self.logger.info('Setting')
        for key, value in self.input.items():
            self.parent.units[key] = [float(value[0].get()), value[1].get()]
        self.win.destroy()
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()
